@@ -9,7 +9,6 @@ from __future__ import annotations
 import typing
 from typing import Any
 from typing import NamedTuple
-from typing import Optional
 from typing import TYPE_CHECKING
 
 import logging
@@ -43,7 +42,8 @@ from nbxmpp.const import StatusCode
 from nbxmpp.language import LanguageMap
 from nbxmpp.language import LanguageRange
 from nbxmpp.language import LanguageTag
-from nbxmpp.modules.dataforms import DataForm
+from nbxmpp.modules.dataforms import MultipleDataForm
+from nbxmpp.modules.dataforms import SimpleDataForm
 from nbxmpp.modules.fallback import FallbacksForT
 from nbxmpp.modules.fallback import strip_fallback
 from nbxmpp.namespaces import Namespace
@@ -133,7 +133,7 @@ class MucDestroyed(NamedTuple):
 
 class MucConfigResult(NamedTuple):
     jid: JID
-    form: Optional[Any] = None
+    form: Any | None = None
 
 
 class MucSubject(NamedTuple):
@@ -169,7 +169,7 @@ class PubSubEventData(NamedTuple):
     node: str
     id: str | None = None
     item: Node | None = None
-    data: Optional[Any] = None
+    data: Any | None = None
     deleted: bool = False
     retracted: bool = False
     purged: bool = False
@@ -330,8 +330,8 @@ class DiscoItem(NamedTuple):
 
 class RegisterData(NamedTuple):
     instructions: str | None
-    form: Optional[Any]
-    fields_form: Optional[Any]
+    form: Any | None
+    fields_form: Any | None
     oob_url: str | None
     bob_data: BobData | None
 
@@ -451,7 +451,7 @@ class DiscoInfo(NamedTuple):
     stanza: Iq | None
     identities: list[DiscoIdentity]
     features: list[str]
-    dataforms: list[DataForm]
+    dataforms: list[SimpleDataForm | MultipleDataForm]
     timestamp: float | None = None
 
     def get_caps_hash(self) -> str | None:
@@ -472,13 +472,15 @@ class DiscoInfo(NamedTuple):
                 continue
         return False
 
-    def get_field_value(self, form_type: str, var: str) -> Optional[Any]:
+    def get_field_value(self, form_type: str, var: str) -> Any | None:
         for dataform in self.dataforms:
             try:
                 if dataform["FORM_TYPE"].value != form_type:
                     continue
 
-                if dataform[var].type_ == "jid-multi":
+                if dataform[var].type_ == "jid-multi" or dataform[var].type_.startswith(
+                    "list-"
+                ):
                     return dataform[var].values or None
                 return dataform[var].value
 
@@ -587,31 +589,31 @@ class DiscoInfo(NamedTuple):
         return None
 
     @property
-    def muc_room_name(self) -> Optional[Any]:
+    def muc_room_name(self) -> Any | None:
         return self.get_field_value(Namespace.MUC_INFO, "muc#roomconfig_roomname")
 
     @property
-    def muc_description(self) -> Optional[Any]:
+    def muc_description(self) -> Any | None:
         return self.get_field_value(Namespace.MUC_INFO, "muc#roominfo_description")
 
     @property
-    def muc_log_uri(self) -> Optional[Any]:
+    def muc_log_uri(self) -> Any | None:
         return self.get_field_value(Namespace.MUC_INFO, "muc#roominfo_logs")
 
     @property
-    def muc_users(self) -> Optional[Any]:
+    def muc_users(self) -> Any | None:
         return self.get_field_value(Namespace.MUC_INFO, "muc#roominfo_occupants")
 
     @property
-    def muc_contacts(self) -> Optional[Any]:
+    def muc_contacts(self) -> Any | None:
         return self.get_field_value(Namespace.MUC_INFO, "muc#roominfo_contactjid")
 
     @property
-    def muc_subject(self) -> Optional[Any]:
+    def muc_subject(self) -> Any | None:
         return self.get_field_value(Namespace.MUC_INFO, "muc#roominfo_subject")
 
     @property
-    def muc_subjectmod(self) -> Optional[Any]:
+    def muc_subjectmod(self) -> Any | None:
         # muc#roominfo_changesubject stems from a wrong example in the MUC XEP
         # Ejabberd and Prosody use this value
         # muc#roomconfig_changesubject is also used by Prosody
@@ -622,7 +624,7 @@ class DiscoInfo(NamedTuple):
         )
 
     @property
-    def muc_lang(self) -> Optional[Any]:
+    def muc_lang(self) -> Any | None:
         return self.get_field_value(Namespace.MUC_INFO, "muc#roominfo_lang")
 
     @property
@@ -723,6 +725,23 @@ class DiscoInfo(NamedTuple):
         except Exception:
             return None
 
+    @property
+    def reactions_per_user(self) -> int | None:
+        value = self.get_field_value(
+            Namespace.REACTIONS_RESTRICTIONS, "max_reactions_per_user"
+        )
+        if value is None:
+            return None
+        try:
+            return int(value)
+        except ValueError:
+            return None
+
+    @property
+    def reactions_allowlist(self) -> list[str] | None:
+        value = self.get_field_value(Namespace.REACTIONS_RESTRICTIONS, "allowlist")
+        return value
+
 
 class DiscoIdentity(NamedTuple):
 
@@ -1011,6 +1030,7 @@ class ReceiptData(NamedTuple):
 class Hat:
     uri: str
     title: str
+    hue: float | None = None
 
 
 class HatData:
@@ -1051,6 +1071,16 @@ class EncryptionData:
     trust: int
 
 
+@dataclass
+class OpenGraphData:
+    title: str | None = None
+    description: str | None = None
+    url: str | None = None
+    image: str | None = None
+    type: str | None = None
+    site_name: str | None = None
+
+
 class Properties:
     pass
 
@@ -1112,6 +1142,9 @@ class MessageProperties:
     security_label: SecurityLabel | None = None
     chatstate: Chatstate | None = None
     reactions: Reactions | None = None
+    # keys of the open_graph dict are URLs defined by the "about" attr of the
+    # corresponding <Description> element (RDF namespace)
+    open_graph: dict[str, OpenGraphData] = field(default_factory=dict)
 
     def is_from_us(self, bare_match: bool = True) -> bool:
         if self.from_ is None:
@@ -1253,7 +1286,7 @@ class IqProperties:
     type: IqType | None = None
     jid: JID | None = None
     id: str | None = None
-    error: Optional[Any] = None
+    error: Any | None = None
     query: Node | None = None
     payload: Node | None = None
     http_auth: HTTPAuthData | None = None
@@ -1283,7 +1316,7 @@ class IqPropertiesBase(typing.Protocol):
     type: IqType
     jid: JID
     id: str
-    error: Optional[Any]
+    error: Any | None
     query: Node | None
     payload: Node | None
 
@@ -1314,8 +1347,8 @@ class PresenceProperties:
     timestamp: float = field(default_factory=time.time)
     user_timestamp: float | None = None
     idle_timestamp: float | None = None
-    signed: Optional[Any] = None
-    error: Optional[Any] = None
+    signed: Any | None = None
+    error: Any | None = None
     avatar_sha: str | None = None
     avatar_state: AvatarState = AvatarState.IGNORE
     muc_jid: JID | None = None
